package WeightedKmeans;

import java.io.BufferedReader;
import java.io.BufferedWriter;
import java.io.FileReader;
import java.io.FileWriter;
import java.util.ArrayList;
import java.util.List;

/**
 * @author Ahmad
 * 
 * 
 */
public class KMeans {

	int k;
	Vector[] patterns;
    List<String> featureNames;
    List<String> encodedFeatures;
	MembershipMatrix mm;
	List<Double> weightList;

	public KMeans(int k, Vector... patterns) {
		this.k = k;
		this.patterns = patterns;
		mm = new MembershipMatrix(patterns.length, k);
		
	}
	
	public KMeans(int k, double[][] inputMatrix, List<Double> weights){
		this.weightList = weights;
		this.k =k;
		
		if(k > inputMatrix.length)
		{
			System.out.println("cluster size cannot be greater than the size of database!!");
		}
		int numberOfPatterns = inputMatrix.length;
		int numberOfFeatures = inputMatrix[0].length;
		patterns = new Vector[numberOfPatterns];
		for(int row = 0; row<numberOfPatterns; row++){
			patterns[row] = new Vector(inputMatrix[row]);	
		}
		mm = new MembershipMatrix(patterns.length, k);
	}

	/**
	 * 
	 * @param k
	 *            - number of clusters
	 * @param dims
	 *            - number of feature in the pattern vectors
	 * @param filename
	 *            - the name of csv contain the data that you wish to cluster
	 * @param WeightListFile
	 *            - the file containing each weight of each feature
	 * @param delimiter
	 *            - the separator for each file.
	 */

	public KMeans(int k, int dims, String filename, String weightListFile,
			String delimiter) {
		this.k = k;
		// Reads in the data
		try {
			BufferedReader in = new BufferedReader(new FileReader(filename));
			List<Vector> patternList = new ArrayList<Vector>();
			for (String line = in.readLine(); line != null; line = in
					.readLine()) {
				String[] strValues = line.split(delimiter);
				double[] values = new double[dims];
				for (int i = 0; i < dims && i < strValues.length; i++) {
					values[i] = Double.valueOf(strValues[i]);
				}
				patternList.add(new Vector(values));
			}
			patterns = new Vector[patternList.size()];
			for (int i = 0; i < patternList.size(); i++) {
				patterns[i] = patternList.get(i);
			}

		} catch (Exception e) {
			e.printStackTrace();
		}
		// Reads in the weights
		try {
			BufferedReader inWeights = new BufferedReader(new FileReader(weightListFile));
			weightList = new ArrayList<Double>();
			for (String line = inWeights.readLine(); line != null; line = inWeights
					.readLine()) {
					weightList.add(Double.valueOf(line));
			}

		} catch (Exception e) {
			e.printStackTrace();
		}

		mm = new MembershipMatrix(patterns.length, k);
	}

	/**
	 * returns an array of integers, the ith elements represents the cluster
	 * index of the ith pattern
	 */
	public List<Integer> partition(int maxRuns) {
		List<Vector> newZ = new ArrayList<Vector>();
		for (int i = 0; i < k; i++) {
			newZ.add(patterns[i]);
		}

		List<Vector> oldZ;

		//This loop continues for MaxRuns ideally it should continue until convergence
		for (int m = 0;m<maxRuns; m++) {
			oldZ = newZ;
			//System.out.println(m);
			// for each pattern, find the cluster center that is closest
			// to that pattern, and then put that pattern into that cluster
			int[] clusterIndexes = new int[patterns.length];
			for (int i = 0; i < patterns.length; i++) {
				int nearestCluster = patterns[i].getNearestPointIndex(newZ,weightList);
				clusterIndexes[i] = nearestCluster;
			}
			mm.moveAllPatterns(clusterIndexes);

			newZ = calculateZ();

			if (newZ.equals(oldZ))
				break;
		}

		return mm.getClusters();
	}

	/**
	 * calculates cluster centers
	 * 
	 * @return
	 */
	public List<Vector> calculateZ() {
		List<Vector> z = new ArrayList<Vector>();
		for (int j = 0; j < k; j++) {
			int[] patternIndexes = mm.getPatterns(j);
			if(patternIndexes.length!=0){
				z.add(Vector.calculateCenter(getPatternsWithIndexes(patternIndexes)));
			}
			
		}

		return z;
	}

	private Vector[] getPatternsWithIndexes(int[] indexes) {
		Vector[] patterns = new Vector[indexes.length];
		for (int i = 0; i < patterns.length; i++) {
			patterns[i] = this.patterns[indexes[i]];
		}

		return patterns;
	}

	public String getResultsString() {

		
		StringBuilder sb = new StringBuilder("");

		List<Integer> clusters = mm.getClusters();
		for (int i = 0; i < patterns.length; i++) {
			sb.append((clusters.get(i) + 1)+ "," +patterns[i].toString()  + "\n");
		}

		return sb.toString();
	}
	
	public double[][] convertVectorArray(List<Vector> patternsIn){
		int lenghtOfPatterns = patternsIn.get(0).size();
		int numberOfPatterns = patternsIn.size();
		double[][] newArray = new double[numberOfPatterns][lenghtOfPatterns];
		for(int row = 0; row<numberOfPatterns; row++){
			for(int column =0; column<lenghtOfPatterns; column++){
				newArray[row][column] = patternsIn.get(row).getValueAtIndex(column);
			}
		}
		
		return newArray;
	}
	
	
	

	public static void main(String[] args) {
		
		String datalocation = "/scratch/Research/DataPrivacyProject/WeightedKmeans/durbodaxTop.dat";
		String weightslocation ="/scratch/Research/DataPrivacyProject/WeightedKmeans/durbodaxWeights.dat";
		KMeans kmeans = new KMeans(80, 62, datalocation, weightslocation, ",");
		int maxRuns =100;
		
			List<Integer> clusters = kmeans.partition(maxRuns);
			List<Vector> centroids = kmeans.calculateZ();
			
			try{
			  // Create file 
			  FileWriter fstream = new FileWriter("/scratch/Research/DataPrivacyProject/WeightedKmeans/clusterResults.txt");
			  BufferedWriter out = new BufferedWriter(fstream);
			  out.write(kmeans.getResultsString());
			  //Close the output stream
			  out.close();
			  }catch (Exception e){//Catch exception if any
				  System.err.println("Error: " + e.getMessage());
			  }
			
			//This section prints the vectors
			
			try{
				  // Create file 
				  FileWriter fstreamCentroids = new FileWriter("/scratch/Research/DataPrivacyProject/WeightedKmeans/centriods.txt");
				  BufferedWriter outCentroids = new BufferedWriter(fstreamCentroids);
				  int i = 1;
				  for(Vector centroid : centroids){
					  outCentroids.write(i+++","+centroid.toString()+"\n");
				  }
				  //Close the output stream
				  outCentroids.close();
				  }catch (Exception e){//Catch exception if any
					  System.err.println("Error: " + e.getMessage());
				  }
			
			
		
	}
}