
package associationRuleMining;

import java.io.*;
import java.util.*;

/**
 * This class represents the whole dataset of all supermarket transactions. A
 * transaction is an itemset of items bought by a supermarket client in a single
 * transaction. The class includes functions for loading the dataset from the
 * file, computing support and confidence etc. The <code> main() </code>
 * function of the class loads the dataset from the default file, runs the
 * apriori algorithm and dumps the results to the console.
 * 
 * 
 */

public class Dataset {

	private LinkedList<Itemset> transactionList = new LinkedList<Itemset>();
	
	/**
	 * Creates and initializes the dataset with data from double matrix of dummy encoded values. 
	 * 
	 * @param encode matris this is the dummy encoded matri
	 * 
	 */
	
	public Dataset(double[][] encoded, boolean[] nominal){
		int rows = encoded.length;
		int cols = encoded[0].length;
		for(int r=0; r<rows; r++){
			Itemset newItemset = new Itemset();
			for(int c=0; c<cols; c++){
				if(nominal[c]){
					newItemset.addItem(new Item(c,encoded[r][c]));
				}
			}
			if (newItemset.size() != 0) {
				transactionList.add(newItemset);
			}
		}
		
	}
	
	


	public void dumpItemsets() {
		Iterator itItemset = getTransactionIterator();
		while (itItemset.hasNext()) {
			Itemset itemset = (Itemset) itItemset.next();
			System.out.println(itemset.toString());
		}
	}

	/**
	 * 
	 * @return the iterator that allows to go over all the transactions in the
	 *         dataset The transactions are <code> Itemset </code> objects
	 */
	public Iterator getTransactionIterator() {
		return transactionList.iterator();
	}

	/**
	 * 
	 * @return the number of transactions in the dataset
	 */
	public int getNumTransactions() {
		return transactionList.size();
	}

	/**
	 * 
	 * @param itemset
	 * @return the support value for a given itemset in the context of the
	 *         current dataset
	 */
	public double computeSupportForItemset(Itemset itemset) {
		int occurrenceCount = 0;
		Iterator itItemset = getTransactionIterator();
		while (itItemset.hasNext()) {
			Itemset shoppingList = (Itemset) itItemset.next();
			if (shoppingList.intersectWith(itemset).size() == itemset.size()) {
				occurrenceCount++;
			}
		}
		return ((double) occurrenceCount) / getNumTransactions();
	}
	
	
	public boolean computeSupportForItemset(Itemset itemset, double support) {
		int occurrenceCount = 0;
		double stopCondition = getNumTransactions()*support;
		Iterator itItemset = getTransactionIterator();
		while (itItemset.hasNext()) {
			Itemset shoppingList = (Itemset) itItemset.next();
			if (shoppingList.intersectWith(itemset).size() == itemset.size()) {
				occurrenceCount++;
				if(occurrenceCount>stopCondition){
					return true;
				}
			}
		}
		return false;
	}


	/**
	 * 
	 * @param associationRule
	 * @return the confidence value for a given association rule in the context
	 *         of the current dataset
	 */
	public double computeConfidenceForAssociationRule(
			AssociationRule associationRule) {
		Itemset union = associationRule.getItemsetA().unionWith(
				associationRule.getItemsetB());
		return computeSupportForItemset(union)
				/ computeSupportForItemset(associationRule.getItemsetA());
	}

	/**
	 * 
	 * @return all possible itemsets of size one based on the current dataset
	 */
	public Set getAllItemsetsOfSizeOne() {
		Iterator itItemset = getTransactionIterator();
		Itemset bigUnion = new Itemset();
		while (itItemset.hasNext()) {
			Itemset itemset = (Itemset) itItemset.next();
			bigUnion = bigUnion.unionWith(itemset);
		}

		// break up the big unioned itemset into one element itemsets
		HashSet allItemsets = new HashSet();
		Iterator itItem = bigUnion.getItemIterator();
		while (itItem.hasNext()) {
			Item item = (Item) itItem.next();
			Itemset itemset = new Itemset();
			itemset.addItem(item);
			allItemsets.add(itemset);
		}

		return allItemsets;
	}

	/**
	 * The core of the association rule mining algorithm. This is what needs to
	 * be implemented. This is the only piece of code that you need to modify to
	 * complete the exercise.
	 *  
	 * @param minSupport
	 *            minimal support value below which itemsets should not be
	 *            considered when generating candidate itemsets
	 * @param minConfidence
	 *            minimal support value for the association rules output by the
	 *            algorithm
	 * @return a collection of <code> AssociationRule </code> instances
	 */
	public Collection runApriori(double minSupport, double minConfidence, int maxCandiates) {
		Collection discoveredAssociationRules = new LinkedList();

		int currentCandiates =0;
		// generate candidate itemsets
		final int MAX_NUM_ITEMS = 2; // All rules must contain less than 3 items;
		Set[] candidates = new Set[MAX_NUM_ITEMS];
		candidates[0] = getAllItemsetsOfSizeOne();
		for (int numItems = 1; numItems < MAX_NUM_ITEMS
				&& !candidates[numItems-1].isEmpty() && currentCandiates<maxCandiates; numItems++) {
			candidates[numItems] = new HashSet();
			for (Iterator itItemset1 = candidates[numItems-1].iterator(); currentCandiates<maxCandiates && itItemset1
					.hasNext();) {
				Itemset itemset1 = (Itemset) itItemset1.next();
				for (Iterator itItemset2 = candidates[numItems-1].iterator(); currentCandiates<maxCandiates && itItemset2
						.hasNext();) {
					Itemset itemset2 = (Itemset) itItemset2.next();
					currentCandiates++; //Attempt
					//System.out.println(currentCandiates);
					if (itemset1.intersectWith(itemset2).size() == numItems-1) {
						Itemset candidateItemset = itemset1.unionWith(itemset2);
						assert (candidateItemset.size() == numItems + 1);
						//System.out.println("currentAttempt: "+currentCandiates);
						if (computeSupportForItemset(candidateItemset, minSupport)) {
							candidates[numItems].add(candidateItemset);
						}
					}
				}
			}
		}

		// generate association rules from candidate itemsets
		for (int numItems = 1; numItems < MAX_NUM_ITEMS
				&& !candidates[numItems].isEmpty(); numItems++) {
			for (Iterator itItemsetCandidate = candidates[numItems].iterator(); itItemsetCandidate
					.hasNext();) {
				Itemset itemsetCandidate = (Itemset) itItemsetCandidate.next();
				for (Iterator itItemsetSub = itemsetCandidate
						.generateAllNonEmptySubsets().iterator(); itItemsetSub
						.hasNext();) {
					Itemset itemsetSub = (Itemset) itItemsetSub.next();
					Itemset itemsetA = itemsetSub;
					Itemset itemsetB = itemsetCandidate.minusAllIn(itemsetSub);
					AssociationRule candidateAssociationRule = new AssociationRule(
							itemsetA, itemsetB);
					if (computeConfidenceForAssociationRule(candidateAssociationRule) > minConfidence) {
						discoveredAssociationRules
								.add(candidateAssociationRule);

					}
				}
			}
		}
		return discoveredAssociationRules;
	}
	
	
	public static double[][] genRanMatrix(int rows, int cols){
		double[][] genMatrix = new double[rows][cols];
		for(int r=0; r<rows; r++){
			for(int c=0; c<cols; c++){
				genMatrix[r][c] = (Math.random()>0.5) ? 1 : 0 ;
			}
		}
		return genMatrix;
		
	}
	
	
	public static boolean[] genRanBool(int cols){
		boolean[] ranBool = new boolean[cols];
		for(int c=0; c<cols; c++){
			ranBool[c] = (Math.random()>0.5) ? true : false ;
		}
		return ranBool;
	}
	
	
	public static void printRules(Collection discoveredAssociationRules, Dataset dataset ){
	Iterator itAssociationRule = discoveredAssociationRules.iterator();
	while (itAssociationRule.hasNext()) {
		AssociationRule associationRule = (AssociationRule) itAssociationRule
				.next();
		System.out
				.println("assoctiation rule: "
						+ associationRule
						+ "\tsupport: "
						+ dataset
								.computeSupportForItemset(associationRule
										.getItemsetA().unionWith(
												associationRule
														.getItemsetB()))
						+ "\tconfidence: "
						+ dataset.computeConfidenceForAssociationRule(associationRule));

	}	
	}
	/**
	 * Loads the dataset from a random boolean matrix, runs the apriori algorithm and outputs the result to the console.
	 * @param args
	 */
	public static void main(String[] args) {
		try {
			
			Dataset dataset = new Dataset(genRanMatrix(10,10),genRanBool(10));
			Collection discoveredAssociationRules = dataset
					.runApriori(0.2, 0.81,10000);

			Iterator itAssociationRule = discoveredAssociationRules.iterator();
			while (itAssociationRule.hasNext()) {
				AssociationRule associationRule = (AssociationRule) itAssociationRule
						.next();
				System.out
						.println("assoctiation rule: "
								+ associationRule
								+ "\tsupport: "
								+ dataset
										.computeSupportForItemset(associationRule
												.getItemsetA().unionWith(
														associationRule
																.getItemsetB()))
								+ "\tconfidence: "
								+ dataset.computeConfidenceForAssociationRule(associationRule));
				
			}

			/*
			 * Itemset itemset = new Itemset(); itemset.addItem(new Item("1"));
			 * itemset.addItem(new Item("2")); itemset.addItem(new Item("3"));
			 * itemset.addItem(new Item("4"));
			 * 
			 * Iterator itItemset = itemset.generateAllNonEmptySubsets()
			 * .iterator(); while (itItemset.hasNext()) { Itemset itemset2 =
			 * (Itemset) itItemset.next();
			 * System.out.println(itemset2.toString()); }
			 */

		} catch (Exception ex) {
			ex.printStackTrace();
		}
	}
}